from cyvlfeat.fisher import fisher
import numpy as np
from numpy.testing import assert_allclose


def test_fisher_dimension():
    N = 1000
    K = 512
    D = 128
    
    x = np.random.uniform(size=(D, N)).astype(np.float32)
    means = np.random.uniform(size=(D, K)).astype(np.float32)
    covariances = np.random.uniform(size=(D, K)).astype(np.float32)
    priors = np.random.uniform(size=(K,)).astype(np.float32)
    priors /= np.linalg.norm(priors)
    enc = fisher(x, means, covariances, priors, verbose=True)
    
    expected = 2 * K * D
    observed = len(enc)
    assert(expected == observed)


def test_fisher_encoding():
    N = 21
    K = 3
    D = 5
    x = np.zeros(D * N, dtype=np.float32)
    for i in range(D * N):
        x[i] = i
    x = x.reshape(D, N)

    mu = np.zeros(D * K, dtype=np.float32)
    for i in range(D * K):
        mu[i] = i
    mu = mu.reshape(D, K)
    
    sigma2 = np.ones((D, K), dtype=np.float32)
    prior = (1.0 / K) * np.ones(K, dtype=np.float32)

    observed_enc = fisher(x, mu, sigma2, prior, verbose=True)

    # expected result obtained from running vl_fisher_encode from a C program
    expected_enc = np.array([0.000000000, 0.000000000, 0.000000000,
                             0.000000000, 0.000000000, 0.000000000,
                             0.000000000, 0.000000000, 0.000000000,
                             0.000000000, 70.519210815, 70.519210815,
                             70.519210815, 70.519210815, 70.519210815,
                             -0.058321182, -0.058321182, -0.058321182,
                             -0.058321182, -0.058321182, -0.058321182,
                             -0.058321182, -0.058321182, -0.058321182,
                             -0.058321182, 3073.876220703, 3073.876220703,
                             3073.876220703, 3073.876220703, 3073.876220703],
                            dtype=np.float32)
    assert_allclose(expected_enc, observed_enc)



from __future__ import division
from cyvlfeat.kmeans import kmeans, kmeans_quantize, ikmeans, ikmeans_push, hikmeans, hikmeans_push
import numpy as np


def set_distance(values1, values2):
    """
    Compare two sets of values and returns the maximum distance between the closest pairs
    Parameters
    ----------
    values1 : NxD set of values
    values2 : MxD set of values

    Returns
    -------
    Distance
    """
    dimension = values1.shape[1]
    assert values2.shape[1] == dimension
    dist = np.sqrt(np.sum(
        np.power(values1.reshape((-1, 1, dimension)) - values2.reshape((1, -1, dimension)), 2),
        axis=2))*1/dimension
    assert isinstance(dist, np.ndarray)
    assert dist.shape == (values1.shape[0], values2.shape[0])
    return max(np.max(np.min(dist, axis=0)), np.max(np.min(dist, axis=1)))


def test_kmeans_float():
    num_data = 50
    num_centers = 4
    dimension = 8
    noise_level = 0.1

    centers = np.random.randint(-40, 40, (num_centers, dimension)).astype(np.float32)
    data = np.empty((num_data, dimension), dtype=np.float32)
    for i in range(num_data):
        data[i] = centers[i % num_centers] + np.random.random_sample(dimension)*noise_level

    found_centers = kmeans(data, num_centers, initialization="PLUSPLUS")
    found_assignments = kmeans_quantize(data, found_centers)

    assert found_centers.dtype == np.float32
    assert found_centers.shape == (num_centers, dimension)

    assert found_assignments.dtype == np.uint32
    assert found_assignments.shape == (num_data,)

    dist = set_distance(centers, found_centers)
    assert dist <= noise_level, dist

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]
    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]


def test_kmeans_double():
    num_data = 50
    num_centers = 4
    dimension = 8
    noise_level = 0.1

    centers = np.random.randint(-40, 40, (num_centers, dimension)).astype(np.float64)
    data = np.empty((num_data, dimension), dtype=np.float64)
    for i in range(num_data):
        data[i] = centers[i % num_centers] + np.random.random_sample(dimension)*noise_level

    found_centers = kmeans(data, num_centers, initialization="PLUSPLUS")
    found_assignments = kmeans_quantize(data, found_centers)

    assert found_centers.dtype == np.float64
    assert found_centers.shape == (num_centers, dimension)

    assert found_assignments.dtype == np.uint32
    assert found_assignments.shape == (num_data,)

    dist = set_distance(centers, found_centers)
    assert dist <= noise_level, dist

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]
    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]


def test_kmeans_ANN():
    num_data = 5000
    num_centers = 4
    dimension = 8
    noise_level = 0.1

    centers = np.random.randint(-40, 40, (num_centers, dimension)).astype(np.float32)
    data = np.empty((num_data, dimension), dtype=np.float32)
    for i in range(num_data):
        data[i] = centers[i % num_centers] + np.random.random_sample(dimension)*noise_level

    found_centers = kmeans(data, num_centers, initialization="PLUSPLUS", algorithm="ANN")
    found_assignments = kmeans_quantize(data, found_centers, algorithm="ANN")

    assert found_centers.dtype == np.float32
    assert found_centers.shape == (num_centers, dimension)

    dist = set_distance(centers, found_centers)
    assert dist <= noise_level, dist

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]
    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]


def test_ikmeans():
    num_data = 5000
    num_centers = 40
    dimension = 128
    noise_level = 3

    centers = np.random.randint(0, 200, (num_centers, dimension)).astype(np.uint8)
    data = np.empty((num_data, dimension), dtype=np.uint8)
    for i in range(num_data):
        data[i] = centers[i % num_centers]
    data = data + np.random.randint(0, noise_level, (num_data, dimension)).astype(np.uint8)

    found_centers, found_assignments = ikmeans(data, num_centers)

    assert found_centers.dtype == np.int32
    assert found_centers.shape == (num_centers, dimension)

    assert found_assignments.dtype == np.uint32
    assert found_assignments.shape == (num_data,)

    # Because the initialization is random, these tests does not work all the time. Two clusters might be merged.
    # dist = set_distance(centers.astype(np.float32), found_centers.astype(np.float32))
    # assert dist <= noise_level, dist

    # for i in range(num_centers):
    #     for j in range(num_centers):
    #         if i != j:
    #             assert found_assignments[i] != found_assignments[j]

    # for i in range(num_data):
    #     assert found_assignments[i] == found_assignments[i % num_centers]

    assignments_2 = ikmeans_push(data, found_centers)
    assert np.allclose(found_assignments, assignments_2)


def test_ikmeans_2():
    num_data = 5000
    num_centers = 2
    dimension = 2
    noise_level = 3

    centers = np.array([[0, 0], [50, 100]], dtype=np.uint8)
    data = np.empty((num_data, dimension), dtype=np.uint8)
    for i in range(num_data):
        data[i] = centers[i % num_centers]
    data = data + np.random.randint(0, noise_level, (num_data, dimension)).astype(np.uint8)

    found_centers, found_assignments = ikmeans(data, num_centers)

    assert found_centers.dtype == np.int32
    assert found_centers.shape == (num_centers, dimension)

    assert found_assignments.dtype == np.uint32
    assert found_assignments.shape == (num_data,)

    dist = set_distance(centers.astype(np.float32), found_centers.astype(np.float32))
    assert dist <= noise_level, dist

    for i in range(num_centers):
        for j in range(num_centers):
            if i != j:
                assert found_assignments[i] != found_assignments[j]

    for i in range(num_data):
        assert found_assignments[i] == found_assignments[i % num_centers]

    assignments_2 = ikmeans_push(data, found_centers)
    assert np.allclose(found_assignments, assignments_2)


def test_hikmeans():
    num_data = 5000
    num_centers = 40
    dimension = 128
    noise_level = 3

    centers = np.random.randint(0, 200, (num_centers, dimension)).astype(np.uint8)
    data = np.empty((num_data, dimension), dtype=np.uint8)
    for i in range(num_data):
        data[i] = centers[i % num_centers]
    data = data + np.random.randint(0, noise_level, (num_data, dimension)).astype(np.uint8)

    tree_structure, found_assignments = hikmeans(data, 4, 64)

    assert found_assignments.dtype == np.uint32
    assert found_assignments.shape == (num_data, tree_structure.depth)

    assignments_2 = hikmeans_push(data, tree_structure)
    assert np.allclose(found_assignments, assignments_2)



# -*- coding: utf-8 -*-
# Author: Alexis Mignon <alexis.mignon@probayes.com>


from .host cimport vl_size, vl_type
from .kmeans cimport VlKMeans

cdef extern from "vl/gmm.h":

    ctypedef enum VlGMMInitialization:
        VlGMMKMeans, VlGMMRand, VlGMMCustom

    ctypedef struct VlGMM:
        pass

    VlGMM *vl_gmm_new (vl_type dataType, vl_size dimension,
                        vl_size numComponents)
    void vl_gmm_delete (VlGMM *self)


    double vl_gmm_cluster(VlGMM *self, void *data, vl_size numData)

    void vl_gmm_set_means(VlGMM *self,  void *means)
    void vl_gmm_set_covariances(VlGMM *self, void* covariances)
    void vl_gmm_set_priors(VlGMM *self, void* priors)

    void vl_gmm_set_num_repetitions(VlGMM *self, vl_size numRepetitions)
    void vl_gmm_set_max_num_iterations(VlGMM *self, vl_size maxNumIterations)
    void vl_gmm_set_verbosity(VlGMM *self, int verbosity)
    void vl_gmm_set_initialization(VlGMM *self, VlGMMInitialization init)
    void vl_gmm_set_kmeans_init_object(VlGMM *self, VlKMeans *kmeans)
    void vl_gmm_set_covariance_lower_bounds(VlGMM *self, double* bounds)
    void vl_gmm_set_covariance_lower_bound(VlGMM *self, double bound)

    void* vl_gmm_get_means(VlGMM *self)
    void* vl_gmm_get_covariances(VlGMM *self)
    void* vl_gmm_get_priors(VlGMM *self)
    void* vl_gmm_get_posteriors(VlGMM *self)

    vl_size vl_gmm_get_num_clusters(VlGMM *self)
    vl_type vl_gmm_get_data_type(VlGMM *self)
    vl_size vl_gmm_get_max_num_iterations(VlGMM *self)
    vl_size vl_gmm_get_num_repetitions(VlGMM *self)
    VlGMMInitialization vl_gmm_get_initialization(VlGMM *self)

    double* vl_gmm_get_covariance_lower_bounds(VlGMM *self)



# Copyright(C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from cyvlfeat._vl.host cimport vl_size
from cyvlfeat._vl.host cimport vl_type

cdef extern from "vl/fisher.h":
    vl_size vl_fisher_encode(void* enc,
                             vl_type dataType,
                             void* means,
                             vl_size dimension,
                             vl_size numClusters,
                             void* covariances,
                             void* priors,
                             void* data,
                             vl_size numData,
                             int flags)
    cdef int VL_FISHER_FLAG_SQUARE_ROOT "VL_FISHER_FLAG_SQUARE_ROOT"
    cdef int VL_FISHER_FLAG_NORMALIZED "VL_FISHER_FLAG_NORMALIZED"
    cdef int VL_FISHER_FLAG_IMPROVED "VL_FISHER_FLAG_IMPROVED"
    cdef int VL_FISHER_FLAG_FAST "VL_FISHER_FLAG_FAST"



# distutils: include_dirs = cyvlfeat
cimport cython
cimport numpy as np


cdef extern from "cy_util.h":
    void py_printf(const char *format, ...)
    void set_python_vl_printf()


cdef inline np.dtype dtype_from_memoryview(cython.view.memoryview arr):
    return np.dtype(arr.view.format)



# Author: Alexis Mignon <alexis.mignon@probayes.com>
from .cygmm import cy_gmm
import numpy as np


def gmm(X, n_clusters=10, max_num_iterations=100, covariance_bound=None,
        init_mode='rand', init_priors=None, init_means=None, init_covars=None,
        n_repetitions=1, verbose=False):
    """Fit a Gaussian mixture model

    Parameters
    ----------
    X : [n_samples, n_features] `float32/float64` `ndarray`
        The data to be fit. One data point per row.
    n_clusters : `int`, optional
        Number of output clusters.
    max_num_iterations : `int`, optional
        The maximum number of EM iterations.
    covariance_bound : `float` or `ndarray`, optional
        A lower bound on the value of the covariance. If a float is given
        then the same value is given for all features/dimensions. If an
        array is given it should have shape [n_features] and give the
        lower bound for each feature.
    init_mode: {'rand', 'kmeans', 'custom'}, optional
        The initialization mode:

          - rand: Initial mean positions are randomly  chosen among
                  data samples
          - kmeans: The K-Means algorithm is used to initialize the cluster
                    means
          - custom: The intial parameters are provided by the user, through
                    the use of ``init_priors``, ``init_means`` and
                    ``init_covars``. Note that if those arguments are given
                    then the ``init_mode`` value is always considered as
                    ``custom``

    init_priors : [n_clusters,] `ndarray`, optional
        The initial prior probabilities on each components
    init_means : [n_clusters, n_features] `ndarray`, optional
        The initial component means.
    init_covars : [n_clusters, n_features] `ndarray`, optional
        The initial diagonal values of the covariances for each component.
    n_repetitions : `int`, optional
        The number of times the fit is performed. The fit with the highest
        likelihood is kept.
    verbose : `bool`, optional
        If ``True``, display information about computing the mixture model.

    Returns
    -------
    means : [n_clusters, n_features] `ndarray`
        The means of the components
    covars : [n_clusters, n_features] `ndarray`
        The diagonal elements of the covariance matrix for each component.
    priors : [n_clusters] `ndarray`
        The prior probability of each component
    ll : `float`
        The found log-likelihood of the input data w.r.t the fitted model
    posteriors : [n_samples, n_clusters] `ndarray`
        The posterior probability of each cluster w.r.t each data points.
    """
    n_samples = X.shape[0]
    n_features = X.shape[1]

    if X.shape[0] == 0:
        raise ValueError('X should contain at least one row')
    if np.isnan(X).any() or np.isinf(X).any():
        raise ValueError("X contains Nans or Infs.")

    if n_clusters <= 0 or n_clusters > n_samples:
        raise ValueError(
            'n_clusters {} must be a positive integer smaller than the '
            'number of data points {}'.format(n_clusters, n_samples)
        )

    if max_num_iterations < 0:
        raise ValueError('max_num_iterations must be non negative')
    if n_repetitions <= 0:
        raise ValueError('n_repetitions must be a positive integer')
    if init_mode not in {'rand', 'custom', 'kmeans'}:
        raise ValueError("init_mode must be one of {'rand', 'custom', 'kmeans'")

    # Make sure we have the correct types
    X = np.ascontiguousarray(X)
    if X.dtype not in [np.float32, np.float64]:
        raise ValueError('Input data matrix must be of type float32 or float64')

    if covariance_bound is not None:
        covariance_bound = np.asarray(covariance_bound, dtype=np.float)

    if init_priors is not None:
        init_priors = np.require(init_priors, requirements='C', dtype=X.dtype)
        if init_priors.shape != (n_clusters,):
            raise ValueError('init_priors does not have the correct size')
    if init_means is not None:
        init_means = np.require(init_means, requirements='C', dtype=X.dtype)
        if init_means.shape != (n_clusters, n_features):
            raise ValueError('init_means does not have the correct size')
    if init_covars is not None:
        init_covars = np.require(init_covars, requirements='C', dtype=X.dtype)
        if init_covars.shape != (n_clusters, n_features):
            raise ValueError('init_covars does not have the correct size')

    all_inits = (init_priors, init_means, init_covars)
    if any(all_inits) and not all(all_inits):
        raise ValueError('Either all or none of init_priors, init_means and '
                         'init_covars must be set.')

    if init_mode == "custom" and not all(all_inits):
        raise ValueError('init_mode==custom implies that all initial '
                         'parameters are given')

    return cy_gmm(X, n_clusters, max_num_iterations, init_mode.encode('utf8'),
                  n_repetitions, int(verbose),
                  covariance_bound=covariance_bound, init_priors=init_priors,
                  init_means=init_means, init_covars=init_covars)



import numpy as np
from .cykmeans import cy_kmeans, cy_kmeans_quantize, algorithm_type, initialization_type, distance_type


def kmeans(data, num_centers, distance="l2", initialization="RANDSEL",
           algorithm="LLOYD", num_repetitions=1, num_trees=3,
           max_num_comparisons=100, max_num_iterations=100,
           min_energy_variation=0., verbose=False):
    """
    Clusters ``data`` using the K-means algorithm where ``data`` can be either
    `float32` or `float64`.
    Returns the centers of the clusters as an `ndarray` of the same dtype
    as ``data``. Also returns the cluster assignments.

    Parameters
    ----------
    data : [N, D] `float32/float64` `ndarray`
        Input data to be clustered.
    num_centers : `int`
        Number of clusters to compute
    distance : `str`, optional
        Distance to be used ["l1", "l2"]. Default : "l2"
    initialization : `str`, optional
        Use either random data points "RANDSEL" or k-means++ "PLUSPLUS"
        to initialize the centers. Default : "RANDSEL"
    algorithm : `str`, optional
        One of "LLOYD", "ELKAN", or "ANN". LLOYD is the standard Lloyd
        algorithm (similar to expectation maximisation). ELKAN is a
        faster version of LLOYD using triangular inequalities to cut
        down significantly the number of sample-to-center
        comparisons. ANN is the same as Lloyd, but uses an approximated
        nearest neighbours (ANN) algorithm to accelerate the
        sample-to-center comparisons. The latter is particularly
        suitable for very large problems. Default : "LLOYD"
    num_repetitions : `int`, optional
        Number of time to restart k-means. The solution with minimal
        energy is returned. Default : 1
    num_trees : `int`, optional
        The number of trees int the randomized KD-Tree forest (for "ANN").
    max_num_comparisons : `int`, optional
        Maximum number of sample-to-center comparisons when searching
        for the closest center (for "ANN").
    max_num_iterations : `int`, optional
        Maximum number of iterations allowed for the kmeans algorithm
        to converge (for "ANN").
    min_energy_variation : `float`, optional
        The minimum relative energy variation before the algorithm choose to
        stop.
    verbose : `bool`, optional
        Outputs information of the converging process

    Returns
    -------
    centers : [num_centers, D] `float32/float64` `ndarray`
        Computed clusters centers from the data points.
    """

    assert isinstance(num_centers, int)
    assert isinstance(verbose, bool)
    if data.ndim != 2:
        raise ValueError('Data should be a 2-D matrix')

    if data.dtype not in [np.float32, np.float64]:
        raise ValueError('Data should be float32 or float64')

    if num_centers > data.shape[0]:
        raise ValueError('num_centers should be a positive integer smaller '
                         'than the number of data points')

    distance_b = distance.encode()
    if distance_b not in distance_type.keys():
        raise ValueError('distance field invalid')

    initialization_b = initialization.encode()
    if initialization_b not in initialization_type.keys():
        raise ValueError('initialization field invalid')

    algorithm_b = algorithm.encode()
    if algorithm_b not in algorithm_type.keys():
        raise ValueError('algorithm field invalid')

    if (not isinstance(num_repetitions, int)) or num_repetitions <= 0:
        raise ValueError('num_repetitions should be a positive integer')
    if (not isinstance(num_trees, int)) or num_trees <= 0:
        raise ValueError('num_trees should be a positive integer')
    if (not isinstance(max_num_comparisons, int)) or max_num_comparisons <= 0:
        raise ValueError('max_num_comparisons should be a positive integer')
    if (not isinstance(max_num_iterations, int)) or max_num_iterations <= 0:
        raise ValueError('max_num_iterations should be a positive integer')
    if (not isinstance(min_energy_variation, float)) or min_energy_variation < 0:
        raise ValueError('min_energy_variation should be a positive float')

    return cy_kmeans(data, num_centers, distance_b, initialization_b,
                     algorithm_b, num_repetitions, num_trees,
                     max_num_comparisons, max_num_iterations,
                     min_energy_variation, verbose)


def kmeans_quantize(data, centers, distance="l2", algorithm="LLOYD",
                    num_trees=3, max_num_comparisons=100, verbose=False):
    """
    Project the given data to the corresponding cluster indices.

    Parameters
    ----------
    data : [N, D] `float32/float64` `ndarray`
        Input data to be projected.
    centers : [K, D] `float32/float64` `ndarray`
        Clusters centers (must be same data type as data)
    distance : `str`, optional
        Distance to be used ["l1", "l2"]. Default : "l2"
    algorithm : `str`, optional
        One of "LLOYD", "ELKAN", or "ANN". LLOYD is the standard Lloyd
        algorithm (similar to expectation maximisation). ELKAN is a
        faster version of LLOYD using triangular inequalities to cut
        down significantly the number of sample-to-center
        comparisons. ANN is the same as Lloyd, but uses an approximated
        nearest neighbours (ANN) algorithm to accelerate the
        sample-to-center comparisons. The latter is particularly
        suitable for very large problems.
    num_trees : `int`, optional
        The number of trees int the randomized KD-Tree forest (for "ANN").
    max_num_comparisons : `int`, optional
        Maximum number of sample-to-center comparisons when searching
        for the closest center (for "ANN").
    verbose : `bool`, optional
        Outputs information of the converging process

    Returns
    -------
    assignments : [N,] `uint32` `ndarray`
        Assignments to the clusters centers.
    """
    assert isinstance(data, np.ndarray)
    assert isinstance(centers, np.ndarray)
    assert isinstance(verbose, bool)
    if data.ndim != 2:
        raise ValueError('Data should be a 2-D matrix')

    if data.dtype not in [np.float32, np.float64]:
        raise ValueError('Data should be float32 or float64')

    if centers.dtype != data.dtype or centers.shape[1] != data.shape[1]:
        raise ValueError('Data and centers should be of the same dtype and '
                         'with the same number of columns')

    assert centers.shape[0] > 0
    assert data.shape[0] > 0

    distance_b = distance.encode()
    if distance_b not in distance_type.keys():
        raise ValueError('distance field invalid')

    algorithm_b = algorithm.encode()
    if algorithm_b not in algorithm_type.keys():
        raise ValueError('algorithm field invalid')

    if (not isinstance(num_trees, int)) or num_trees <= 0:
        raise ValueError('num_trees should be a positive integer')
    if (not isinstance(max_num_comparisons, int)) or max_num_comparisons <= 0:
        raise ValueError('max_num_comparisons should be a positive integer')

    return cy_kmeans_quantize(data, centers, distance_b, algorithm_b, num_trees,
                              max_num_comparisons, verbose)



# Author: Alexis Mignon <alexis.mignon@probayes.com>
from cyvlfeat._vl.host cimport *
from cyvlfeat._vl.gmm cimport *
from cyvlfeat.cy_util cimport (dtype_from_memoryview, py_printf,
                               set_python_vl_printf)
from libc.string cimport memcpy
import numpy as np
cimport numpy as np


ctypedef fused floats:
    np.float32_t
    np.float64_t


initialization_type = {
    b'rand': VlGMMRand,
    b'kmeans': VlGMMKMeans,
    b'custom': VlGMMCustom
}

inv_initialization_type = {
    VlGMMRand: b'rand',
    VlGMMKMeans: b'kmeans',
    VlGMMCustom: b'custom'
}


def cy_gmm(floats[:, :] data, int n_clusters, int max_num_iterations,
           bytes init_mode, int num_repetitions, int verbose,
           floats[:] covariance_bound=None, floats[:] init_priors=None,
           floats[:, :] init_means=None, floats[:, :] init_covars=None,):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        vl_size i
        vl_size n_samples = data.shape[0]
        vl_size n_features = data.shape[1]

        floats covar_scalar_bound = np.nan
        floats *covar_arr_bound = NULL

        floats LL = 0
        VlGMMInitialization vl_init_mode

        vl_type vl_data_type
        VlGMM *gmm

        floats[:, ::1] out_means
        floats[:, ::1] out_covars
        floats[::1] out_priors
        floats[:, ::1] out_posteriors

    vl_init_mode = initialization_type[init_mode]

    dtype = dtype_from_memoryview(data)
    if dtype == np.float32:
          vl_data_type = VL_TYPE_FLOAT
    elif dtype == np.float64:
          vl_data_type = VL_TYPE_DOUBLE

    if covariance_bound is not None:
        if covariance_bound.shape[0] == 1:
            covar_scalar_bound = covariance_bound[0]
        else:
            covar_arr_bound = &covariance_bound[0]

    gmm = vl_gmm_new(vl_data_type, n_features, n_clusters)
    vl_gmm_set_verbosity(gmm, verbose)
    vl_gmm_set_num_repetitions(gmm, num_repetitions)
    vl_gmm_set_max_num_iterations(gmm, max_num_iterations)
    vl_gmm_set_initialization(gmm, vl_init_mode)

    if not np.isnan(covar_scalar_bound):
        vl_gmm_set_covariance_lower_bound(gmm, covar_scalar_bound)

    if covar_arr_bound:
        vl_gmm_set_covariance_lower_bounds(gmm, <double*>covar_arr_bound)

    if init_priors is not None:
        vl_gmm_set_priors(gmm, <void*>&init_priors[0])

    if init_means is not None:
        vl_gmm_set_means(gmm, <void*>&init_means[0, 0])

    if init_covars is not None:
        vl_gmm_set_covariances(gmm, <void*>&init_covars[0, 0])

    if verbose > 0:
        mode = inv_initialization_type[vl_gmm_get_initialization(gmm)]
        py_printf("vl_gmm: vl_init_mode = %s\n", <const char *>mode)
        py_printf("vl_gmm: maxNumIterations = %llu\n", vl_gmm_get_max_num_iterations(gmm))
        py_printf("vl_gmm: numRepetitions = %llu\n", vl_gmm_get_num_repetitions(gmm))
        py_printf("vl_gmm: data type = %s\n", vl_get_type_name(vl_gmm_get_data_type(gmm)))
        py_printf("vl_gmm: data n_features = %llu\n", n_features)
        py_printf("vl_gmm: num. data points = %llu\n", n_samples)
        py_printf("vl_gmm: num. Gaussian modes = %d\n", n_clusters)

        py_printf("vl_gmm: lower bound on covariance = [")
        if n_features < 3:
            for i in range(n_features):
                py_printf(" %f", vl_gmm_get_covariance_lower_bounds(gmm)[i])
        else:
            py_printf(" %f %f ... %f",
                      vl_gmm_get_covariance_lower_bounds(gmm)[0],
                      vl_gmm_get_covariance_lower_bounds(gmm)[1],
                      vl_gmm_get_covariance_lower_bounds(gmm)[n_features-1])

        py_printf("%s", "]\n")

    # Clustering .....................................
    LL = vl_gmm_cluster(gmm, <void*>&data[0, 0], n_samples)

    out_means = np.zeros((n_clusters, n_features), dtype=dtype)
    out_covars = np.zeros((n_clusters, n_features), dtype=dtype)
    out_priors = np.zeros(n_clusters, dtype=dtype)
    out_posteriors = np.zeros((n_samples, n_clusters), dtype=dtype)

    memcpy(<void*>&out_means[0, 0],
           vl_gmm_get_means(gmm),
           vl_get_type_size(vl_data_type) * n_features * vl_gmm_get_num_clusters(gmm))

    memcpy(<void*>&out_covars[0, 0],
           vl_gmm_get_covariances(gmm),
           vl_get_type_size(vl_data_type) * n_features * vl_gmm_get_num_clusters(gmm))

    memcpy(<void*>&out_priors[0],
           vl_gmm_get_priors(gmm),
           vl_get_type_size(vl_data_type) * vl_gmm_get_num_clusters(gmm))

    memcpy(<void*>&out_posteriors[0, 0],
           vl_gmm_get_posteriors(gmm),
           vl_get_type_size(vl_data_type) * n_samples * vl_gmm_get_num_clusters(gmm))

    vl_gmm_delete(gmm)

    return (np.asarray(out_means), np.asarray(out_covars),
            np.asarray(out_priors), LL, np.asarray(out_posteriors))



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from cyvlfeat._vl.generic cimport *


cpdef void set_simd_enabled(bint x):
    vl_set_simd_enabled(x)

cpdef bint get_simd_enabled():
    return vl_get_simd_enabled()

cpdef bint cpu_has_avx():
    return vl_cpu_has_avx()

cpdef bint cpu_has_sse3():
    return vl_cpu_has_sse3()

cpdef bint cpu_has_sse2():
    return vl_cpu_has_sse2()

cpdef int get_num_cpus():
    """

    Returns
    -------
        Number of CPU cores.
    """
    return vl_get_num_cpus()

cpdef int get_max_threads():
    """
    This function returns the maximum number of thread used by VLFeat.
    VLFeat will try to use this number of computational threads and never exceed it.

    This is similar to the OpenMP function omp_get_max_threads(); however, it reads
    a parameter private to VLFeat which is independent of the value used by the OpenMP library.

    If VLFeat was compiled without OpenMP support, this function returns 1.

    Returns
    -------
        Number of threads.
    """
    return vl_get_max_threads()

cpdef void set_num_threads(int num_threads):
    """
    This function sets the maximum number of computational threads that will be used by VLFeat.
    VLFeat may in practice use fewer threads (for example because ``num_threads`` is larger than the
    number of computational cores in the host, or because the number of threads exceeds the
    limit available to the application).

    If ``num_threads`` is set to 0, then VLFeat sets the number of threads to the OpenMP current
    maximum, obtained by calling omp_get_max_threads().

    This function is similar to omp_set_num_threads() but changes a parameter internal
    to VLFeat rather than affecting OpenMP global state.

    If VLFeat was compiled without, this function does nothing.
    """
    vl_set_num_threads(num_threads)

cpdef int get_thread_limit():
    """
    This function wraps the OpenMP function omp_get_thread_limit().
    If VLFeat was compiled without OpenMP support, this function returns 1.
    If VLFeat was compiled with OpenMP prior to version 3.0 (2008/05), it returns 0.

    Returns
    -------
        Thread limit.
    """
    return vl_get_thread_limit()



import numpy as np
from .cyhog import cy_hog


def hog(image, cell_size, variant='UoCTTI', n_orientations=9,
        directed_polar_field=False, undirected_polar_field=False,
        bilinear_interpolation=False, verbose=False):
    r"""
    Computes the HOG features for ``image``
    and the specified ``cell_size``. ``image`` can be either greyscale or RGB
    and must be ``float32``.

    HOG decomposes an image into small squared cells, computes an histogram of
    oriented gradients in each cell, normalizes the result using a block-wise
    pattern, and return a descriptor for each cell.

    In this case the feature has 31 dimensions. HOG exists in many variants.
    VLFeat supports two: the UoCTTI variant (used by default) and the original
    Dalal-Triggs variant (with 2x2 square HOG blocks for normalization).
    The main difference is that the UoCTTI variant computes both directed and
    undirected gradients as well as a four dimensional texture-energy feature,
    but projects the result down to 31 dimensions.

    ``output`` is a ``float32`` array, its number
    of columns is approximately the number of columns of ``image`` divided
    by ``cell_size`` and the same for the number of rows. The third
    dimension spans the number of feature components.

    Parameters
    ----------
    image : [H, W] or [H, W, 1] or [H, W, 3] `float32` `ndarray`
        A single channel, RGB or greyscale, `float32` numpy array (ndarray)
        representing the image to calculate descriptors for.
    cell_size : `int`
        The size of the cells that the image should be decomposed into. Assumed
        to be square and this only a single integer is required.
    variant : `str`, optional
        Choose a HOG variant from the set {'UoCTTI', 'DalalTriggs'}
    n_orientations : `int`, optional
        Choose a number of undirected orientations in the orientation
        histograms. The angle ``[0, pi)`` is divided in to ``n_orientations``
        equal parts.
    directed_polar_field : `bool`, optional
        By specifying this flag, the image is interpreted as samples
        from a 2D vector field. Angles are measure clockwise, the y-axis
        pointing downwards, starting from the x-axis (pointing to the right).
        The first channel should be the gradient magnitude and the second
        channel should be the directed (``[0, 2pi)``) orientations.
    undirected_polar_field : `bool`, optional
        By specifying this flag, the image is interpreted as samples
        from a 2D vector field as above, however, the angles will be forcibly
        wrapped into the range ``[0, pi)``.
    bilinear_interpolation : `bool`, optional
        This flags activates the use of bilinear interpolation to assign
        orientations to bins. This produces a smoother feature, but is
        only used for the 'DalalTriggs' variant.
    verbose : `bool`, optional
        If ``True``, be verbose.

    Returns
    -------
    output : `(R, R, D)` `float32` `ndarray`
        ``R`` is is approximately the number of columns of ``image`` divided
        by ``cell_size``. ``D`` is the number of feature dimensions.
    """
    # Add a channels axis
    if image.ndim == 2:
        image = image[..., None]

    # Validate image size
    if image.ndim != 3:
        raise ValueError('Only 2D arrays with a channels axis are supported')

    # Validate all the parameters
    if cell_size < 0:
        raise ValueError('cell_size must be > 0')
    if variant not in {'UoCTTI', 'DalalTriggs'}:
        raise ValueError("variant must be in set {'UoCTTI', 'DalalTriggs'}")
    if n_orientations < 0:
        raise ValueError('n_orientations must be > 0')
    if (directed_polar_field or undirected_polar_field) and image.shape[-1] != 2:
        raise ValueError('Expected a polar field image of n_channels == 2')

    # Ensure types are correct before passing to Cython
    image = np.require(image, dtype=np.float32, requirements='C')

    # Shortcut for getting the correct enum value, since is_UoCTTI has
    # an enum value of 1 (True is 1 in Python)
    is_UoCTTI = variant == 'UoCTTI'

    return cy_hog(image, cell_size, is_UoCTTI,
                  n_orientations,  directed_polar_field,
                  undirected_polar_field, bilinear_interpolation,
                  True,  # Return channels as last axis
                  verbose)



import cyvlfeat.fisher
import cyvlfeat.generic
import cyvlfeat.gmm
import cyvlfeat.hog
import cyvlfeat.kmeans
import cyvlfeat.sift


from ._version import get_versions
__version__ = get_versions()['version']
del get_versions



# Copyright(C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from .host cimport vl_size, vl_uint32, vl_uint8, vl_int32, vl_uint


cdef extern from "vl/ikmeans.h":
    ctypedef vl_int32 vl_ikmacc_t

    cdef enum VlIKMAlgorithms:
        VL_IKM_LLOYD = 0,  # Lloyd algorithm
        VL_IKM_ELKAN = 1   # Elkan algorithm

    cdef struct _VlIKMFilt:
      vl_size M                # data dimensionality
      vl_size K                # number of centers
      vl_size max_niters       # Lloyd: maximum number of iterations
      int method               # Learning method
      int verb                 # verbosity level
      vl_ikmacc_t *centers     # centers
      vl_ikmacc_t *inter_dist  # centers inter-distances
    ctypedef _VlIKMFilt VlIKMFilt
    
    # Create and destroy
    VlIKMFilt* vl_ikm_new (int method)
    void vl_ikm_delete (VlIKMFilt *f)
    
    # Process data
    void vl_ikm_init (VlIKMFilt *f, vl_ikmacc_t *centers, vl_size M, vl_size K)
    void vl_ikm_init_rand (VlIKMFilt *f, vl_size M, vl_size K)
    void vl_ikm_init_rand_data (VlIKMFilt *f, vl_uint8 *data, vl_size M,
                                vl_size N, vl_size K)
    int  vl_ikm_train (VlIKMFilt *f, vl_uint8 *data, vl_size N)
    void vl_ikm_push (VlIKMFilt *f, vl_uint32 *asgn, vl_uint8 *data, vl_size N)
    vl_uint vl_ikm_push_one (vl_ikmacc_t* centers, vl_uint8 *data, vl_size M,
                             vl_size K)

    
    # Retrieve data and parameters
    vl_size vl_ikm_get_ndims (VlIKMFilt *f)
    vl_size vl_ikm_get_K (VlIKMFilt *f)
    int vl_ikm_get_verbosity (VlIKMFilt *f)
    vl_size vl_ikm_get_max_niters (VlIKMFilt *f)
    vl_ikmacc_t* vl_ikm_get_centers (VlIKMFilt *f)
    
    # Set parameters
    void vl_ikm_set_verbosity (VlIKMFilt *f, int verb)
    void vl_ikm_set_max_niters (VlIKMFilt *f, vl_size max_niters)



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

cdef extern from "vl/host.h":
    ctypedef unsigned int vl_type
    ctypedef int vl_bool
    ctypedef long long vl_int64  # @brief Signed 64-bit integer.
    ctypedef int vl_int32  # @brief Signed 32-bit integer.
    ctypedef short vl_int16  # @brief Signed 16-bit integer.
    ctypedef char vl_int8  # @brief Signed  8-bit integer.
    ctypedef unsigned long long vl_uint64  # @brief Unsigned 64-bit integer.
    ctypedef unsigned int vl_uint32  # @brief Unsigned 32-bit integer.
    ctypedef unsigned short vl_uint16  # @brief Unsigned 16-bit integer.
    ctypedef unsigned char vl_uint8  # @brief Unsigned  8-bit integer.
    ctypedef int vl_int
    ctypedef unsigned int vl_uint
    ctypedef unsigned long long vl_size
    ctypedef unsigned long long vl_uindex
    ctypedef long long vl_index
    cdef int VL_TYPE_FLOAT "VL_TYPE_FLOAT"
    cdef int VL_TYPE_DOUBLE "VL_TYPE_DOUBLE"
    inline char* vl_get_type_name(vl_type type)
    inline vl_size vl_get_type_size(vl_type type)
    cdef enum:
        VL_FALSE = 0
    cdef enum:
        VL_TRUE = 1



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
cimport numpy as np
from cyvlfeat._vl.host cimport *
from cyvlfeat._vl.mathop cimport *
from cyvlfeat._vl.kmeans cimport *
from cyvlfeat._vl.ikmeans cimport *
from cyvlfeat._vl.hikmeans cimport *
from cyvlfeat.cy_util cimport (py_printf, dtype_from_memoryview,
                               set_python_vl_printf)
from libc.string cimport memcpy
from libc.stdlib cimport malloc


ctypedef fused floats:
    np.float32_t
    np.float64_t


algorithm_type = {
    b'LLOYD': VlKMeansLloyd,
    b'ELKAN': VlKMeansElkan,
    b'ANN': VlKMeansANN
}

algorithm_type_ikmeans = {
    b'LLOYD': VL_IKM_LLOYD,
    b'ELKAN': VL_IKM_ELKAN
}

initialization_type = {
    b'RANDSEL': VlKMeansRandomSelection,
    b'PLUSPLUS': VlKMeansPlusPlus
}

distance_type = {
    b'l1': VlDistanceL1,
    b'l2': VlDistanceL2
}


cpdef cy_kmeans(floats[:, ::1] data, int num_centers, bytes distance,
                bytes initialization, bytes algorithm, int num_repetitions,
                int num_trees, int max_num_comparisons, int max_num_iterations,
                float min_energy_variation, bint verbose):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        VlKMeans* kmeans
        double energy
        int dimension = data.shape[1]
        int num_data = data.shape[0]
        vl_type vl_data_type
        floats[:, :] centers

    dtype = dtype_from_memoryview(data)
    if dtype == np.float32:
        vl_data_type = VL_TYPE_FLOAT
    else:
        vl_data_type = VL_TYPE_DOUBLE
    kmeans = vl_kmeans_new(vl_data_type, distance_type[distance])

    vl_kmeans_set_verbosity(kmeans, verbose)
    vl_kmeans_set_num_repetitions(kmeans, num_repetitions)
    vl_kmeans_set_algorithm(kmeans, algorithm_type[algorithm])
    vl_kmeans_set_initialization(kmeans, initialization_type[initialization])
    vl_kmeans_set_max_num_iterations(kmeans, max_num_iterations)
    vl_kmeans_set_max_num_comparisons(kmeans, max_num_comparisons)
    vl_kmeans_set_num_trees(kmeans, num_trees)

    if min_energy_variation > 0:
        vl_kmeans_set_min_energy_variation(kmeans, min_energy_variation)
        
    if verbose:
        py_printf("kmeans: Initialization = %s\n", initialization)
        py_printf("kmeans: Algorithm = %s\n", algorithm)
        py_printf("kmeans: MaxNumIterations = %llu\n", vl_kmeans_get_max_num_iterations(kmeans))
        py_printf("kmeans: MinEnergyVariation = %f\n", vl_kmeans_get_min_energy_variation(kmeans))
        py_printf("kmeans: NumRepetitions = %llu\n", vl_kmeans_get_num_repetitions(kmeans))
        py_printf("kmeans: data type = %s\n", vl_get_type_name(vl_kmeans_get_data_type(kmeans)))
        py_printf("kmeans: distance = %s\n", vl_get_vector_comparison_type_name(vl_kmeans_get_distance(kmeans)))
        py_printf("kmeans: data dimension = %d\n", dimension)
        py_printf("kmeans: num. data points = %d\n", num_data)
        py_printf("kmeans: num. centers = %d\n", num_centers)
        py_printf("kmeans: max num. comparisons = %d\n", max_num_comparisons)
        py_printf("kmeans: num. trees = %d\n", num_trees)

    energy = vl_kmeans_cluster(kmeans, <void*>&data[0, 0], dimension,
                               num_data, num_centers)

    centers = np.empty((num_centers, dimension), dtype=dtype, order='C')
    memcpy(<void*>&centers[0, 0], vl_kmeans_get_centers(kmeans),
          vl_get_type_size(vl_data_type) * dimension * vl_kmeans_get_num_centers(kmeans))

    vl_kmeans_delete(kmeans)

    return np.asarray(centers)


cpdef cy_kmeans_quantize(floats[:, ::1] data, floats[:, ::1] centers,
                         bytes distance, bytes algorithm, int num_trees,
                         int max_num_comparisons, bint verbose):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        VlKMeans *kmeans
        vl_size dimension = data.shape[1]
        vl_size num_data = data.shape[0]
        vl_size num_centers = centers.shape[0]
        np.uint32_t[::1] assignments = np.empty((num_data,), dtype=np.uint32,
                                                order='C')

    dtype = dtype_from_memoryview(data)
    kmeans = vl_kmeans_new(VL_TYPE_FLOAT if dtype == np.float32 else VL_TYPE_DOUBLE,
                           distance_type[distance])

    vl_kmeans_set_verbosity(kmeans, verbose)
    vl_kmeans_set_algorithm(kmeans, algorithm_type[algorithm])
    vl_kmeans_set_max_num_comparisons(kmeans, max_num_comparisons)
    vl_kmeans_set_num_trees(kmeans, num_trees)
    vl_kmeans_set_centers(kmeans, <void*>&centers[0, 0], dimension, num_centers)

    if algorithm_type[algorithm] == VlKMeansANN:
        vl_kmeans_quantize_ann(kmeans, &assignments[0], NULL, <void*>&data[0, 0],
                               num_data, 0)
    else:
        vl_kmeans_quantize(kmeans, &assignments[0], NULL, <void*>&data[0, 0],
                           num_data)

    vl_kmeans_delete(kmeans)

    return np.asarray(assignments)


cpdef cy_ikmeans(np.uint8_t[:, ::1] data, int num_centers, bytes algorithm,
                 int max_num_iterations, bint verbose):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        VlIKMFilt *ikmf
        int err
        int M = data.shape[1]
        int N = data.shape[0]
        int K = num_centers
        np.int32_t[:, :] centers = np.empty((K, M), dtype=np.int32, order='C')
        np.uint32_t[:] assignments = np.empty((N,), dtype=np.uint32, order='C')

    ikmf = vl_ikm_new(algorithm_type_ikmeans[algorithm])

    vl_ikm_set_verbosity(ikmf, verbose)
    vl_ikm_set_max_niters(ikmf, max_num_iterations)

    vl_ikm_init_rand_data(ikmf, &data[0, 0], M, N, K)

    err = vl_ikm_train(ikmf, &data[0, 0], N)
    if err:
        py_printf("ikmeans: possible overflow!\n")

    memcpy(&centers[0, 0], vl_ikm_get_centers(ikmf),
           sizeof(vl_ikmacc_t) * M * K)

    vl_ikm_push(ikmf, &assignments[0], &data[0, 0], N)

    vl_ikm_delete(ikmf)

    if verbose:
        py_printf("ikmeans: done\n")

    return np.asarray(centers), np.asarray(assignments)


cpdef cy_ikmeans_push(np.uint8_t[:, ::1] data, np.int32_t[:, ::1] centers):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        VlIKMFilt *ikmf = vl_ikm_new(VL_IKM_LLOYD)
        int M = data.shape[1]
        int N = data.shape[0]
        int K = centers.shape[0]
        np.uint32_t[:] assignments = np.empty((N,), dtype=np.uint32, order='C')

    vl_ikm_set_verbosity(ikmf, 0)
    vl_ikm_init(ikmf, <vl_ikmacc_t*>&centers[0, 0], M, K)
    vl_ikm_push(ikmf, &assignments[0], &data[0, 0], N)
    vl_ikm_delete(ikmf)

    return np.asarray(assignments)


class PyHIKMNode:
    def __init__(self):
        self.centers = None
        self.children = []


class PyHIKMTree(PyHIKMNode):
    def __init__(self, K, depth):
        PyHIKMNode.__init__(self)
        self.K = K
        self.depth = depth


cdef build_py_node(py_node, VlHIKMNode *node):
    cdef:
        np.int32_t[:, ::1] centers
        vl_size node_K, M, k

    node_K = vl_ikm_get_K(node.filter)
    M = vl_ikm_get_ndims(node.filter)

    centers = np.empty((node_K, M), dtype=np.int32, order='C')
    if node_K > 0:
        memcpy(&centers[0, 0], vl_ikm_get_centers(node.filter),
               sizeof(vl_ikmacc_t) * M * node_K)
    py_node.centers = centers

    if node.children:
        for k in range(node_K):
            msub = PyHIKMNode()
            build_py_node(msub, node.children[k])
            py_node.children.append(msub)


cdef hikm_to_python(VlHIKMTree* tree):
    py_tree = PyHIKMTree(vl_hikm_get_K(tree), vl_hikm_get_depth (tree))
    build_py_node(py_tree, tree.root)
    return py_tree


cdef VlHIKMNode* build_vl_node(py_node, VlHIKMTree *tree):
    cdef:
        vl_size M, node_K
        VlHIKMNode *node
        np.int32_t[:, ::1] centers_view

    assert isinstance(py_node, PyHIKMNode)

    M = py_node.centers.shape[1]
    node_K = py_node.centers.shape[0]

    assert node_K <= tree.K
    if tree.M == 0:
        tree.M = M
    else:
        assert M == tree.M

    node = <VlHIKMNode*>malloc(sizeof(VlHIKMNode))
    node.filter = vl_ikm_new(tree.method)
    node.children = NULL

    if node_K > 0:
        centers_view = py_node.centers
        vl_ikm_init(node.filter, <vl_ikmacc_t*>&centers_view[0, 0], M,
                    node_K)
    else:
        vl_ikm_init(node.filter, <vl_ikmacc_t*>NULL, M, node_K)

    if len(py_node.children) > 0:
        assert len(py_node.children) == node_K

        node.children = <VlHIKMNode**>malloc(sizeof(VlHIKMNode*) * node_K)
        for k, child in enumerate(py_node.children):
            node.children[k] = build_vl_node(child, tree)
    return node


cdef VlHIKMTree* python_to_hikm(py_tree, int method_type):
    cdef:
        int K, depth
        VlHIKMTree *tree

    assert isinstance(py_tree, PyHIKMTree), "Tree must be of class PyHIKMTree"
    K = py_tree.K
    depth = py_tree.depth

    tree = <VlHIKMTree*>malloc(sizeof(VlHIKMTree))
    tree.depth = <vl_size>depth
    tree.K = <vl_size>K
    tree.M = 0  # to be initialized later
    tree.method = method_type
    tree.root = build_vl_node(py_tree, tree)
    return tree


cpdef cy_hikmeans(np.uint8_t[:, ::1] data, int num_clusters, int num_leaves,
                  int depth, bytes algorithm, bint verbose):
    cdef:
        VlHIKMTree *tree
        int M = data.shape[1]
        int N = data.shape[0]
        int K = num_clusters
        np.uint32_t[:, ::1] assignments

    tree  = vl_hikm_new(algorithm_type_ikmeans[algorithm])

    if verbose:
        py_printf("hikmeans: # dims: %d\n", M)
        py_printf("hikmeans: # data: %d\n", N)
        py_printf("hikmeans: K: %d\n", K)
        py_printf("hikmeans: depth: %d\n", depth)

    vl_hikm_set_verbosity(tree, verbose)
    vl_hikm_init(tree, M, K, depth)
    vl_hikm_train(tree, &data[0,0], N)

    py_tree = hikm_to_python(tree)

    assignments = np.empty((N, depth), dtype=np.uint32, order='C')
    vl_hikm_push(tree, &assignments[0,0], &data[0,0], N)

    if verbose:
        py_printf("hikmeans: done.\n")

    vl_hikm_delete(tree)

    return py_tree, np.asarray(assignments)


cpdef cy_hikmeans_push(np.uint8_t[:, ::1] data, py_tree, bint verbose):
    cdef:
        VlHIKMTree* tree
        int depth
        int N = data.shape[0]
        np.uint32_t[:, ::1] assignments

    tree = python_to_hikm(py_tree, VL_IKM_LLOYD)
    depth = vl_hikm_get_depth(tree)

    if verbose:
        py_printf("vl_hikmeanspush: ndims: %llu K: %llu depth: %d\n",
                  vl_hikm_get_ndims(tree),
                  vl_hikm_get_K(tree),
                  depth)

    assignments = np.empty((N, depth), dtype=np.uint32, order='C')
    vl_hikm_push(tree, &assignments[0, 0], &data[0, 0], N)

    vl_hikm_delete(tree)

    return assignments



# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# Copyright (C) 2013 Andrea Vedaldi and David Novotny.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
from .host cimport vl_size, vl_index, vl_bool, vl_type, vl_uint32
from .mathop cimport (VlFloatVectorComparisonFunction,
                      VlDoubleVectorComparisonFunction, VlVectorComparisonType)


cdef extern from "vl/kmeans.h":
    
    cdef enum _VlKMeansAlgorithm:
        VlKMeansLloyd = 0,       # Lloyd algorithm
        VlKMeansElkan = 1,       # Elkan algorithm
        VlKMeansANN = 2          # Approximate nearest neighbors
    ctypedef _VlKMeansAlgorithm VlKMeansAlgorithm

    cdef enum _VlKMeansInitialization:
        VlKMeansRandomSelection = 0,  # Randomized selection
        VlKMeansPlusPlus = 1          # Plus plus randomized selection
    ctypedef _VlKMeansInitialization VlKMeansInitialization

    cdef struct _VlKMeans:
        vl_type dataType           # Data type.
        vl_size dimension          # Data dimensionality.
        vl_size numCenters         # Number of centers.
        vl_size numTrees           # Number of trees in forest when using ANN-kmeans.
        vl_size maxNumComparisons  # Maximum number of comparisons when using ANN-kmeans.

        VlKMeansInitialization initialization  # Initialization algorithm.
        VlKMeansAlgorithm algorithm            # Clustering algorithm.
        VlVectorComparisonType distance        # Distance.
        vl_size maxNumIterations               # Maximum number of refinement iterations.
        double minEnergyVariation              # Minimum energy variation.
        vl_size numRepetitions                 # Number of clustering repetitions.
        int verbosity                          # Verbosity level.

        void *centers                          # Centers
        void *centerDistances                  # Centers inter-distances.

        double energy                          # Current solution energy.
        VlFloatVectorComparisonFunction floatVectorComparisonFn
        VlDoubleVectorComparisonFunction doubleVectorComparisonFn
    ctypedef _VlKMeans VlKMeans

    VlKMeans* vl_kmeans_new(vl_type dataType, VlVectorComparisonType distance)
    VlKMeans* vl_kmeans_new_copy(VlKMeans *kmeans)
    void vl_kmeans_delete(VlKMeans *self)


    void vl_kmeans_reset(VlKMeans *self)

    double vl_kmeans_cluster(VlKMeans *self,
                             void *data,
                             vl_size dimension,
                             vl_size numData,
                             vl_size numCenters)

    void vl_kmeans_quantize(VlKMeans *self,
                            vl_uint32 *assignments,
                            void *distances,
                            void *data,
                            vl_size numData)

    void vl_kmeans_quantize_ann(VlKMeans *self,
                                vl_uint32 *assignments,
                                void *distances,
                                void *data,
                                vl_size numData,
                                vl_size iteration )

    void vl_kmeans_set_centers(VlKMeans *self,
                               void *centers,
                               vl_size dimension,
                               vl_size numCenters)

    void vl_kmeans_init_centers_with_rand_data(VlKMeans *self,
                                               void *data,
                                               vl_size dimensions,
                                               vl_size numData,
                                               vl_size numCenters)

    void vl_kmeans_init_centers_plus_plus(VlKMeans *self,
                                          void *data,
                                          vl_size dimensions,
                                          vl_size numData,
                                          vl_size numCenters)

    double vl_kmeans_refine_centers(VlKMeans *self,
                                    void *data,
                                    vl_size numData)

    inline vl_type vl_kmeans_get_data_type(VlKMeans *self)
    inline VlVectorComparisonType vl_kmeans_get_distance(VlKMeans *self)

    inline VlKMeansAlgorithm vl_kmeans_get_algorithm(VlKMeans *self)
    inline VlKMeansInitialization vl_kmeans_get_initialization(VlKMeans *self)
    inline vl_size vl_kmeans_get_num_repetitions(VlKMeans *self)

    inline vl_size vl_kmeans_get_dimension(VlKMeans *self)
    inline vl_size vl_kmeans_get_num_centers(VlKMeans *self)

    inline int vl_kmeans_get_verbosity(VlKMeans *self)
    inline vl_size vl_kmeans_get_max_num_iterations(VlKMeans *self)
    inline double vl_kmeans_get_min_energy_variation(VlKMeans *self)
    inline vl_size vl_kmeans_get_max_num_comparisons(VlKMeans *self)
    inline vl_size vl_kmeans_get_num_trees(VlKMeans *self)
    inline double vl_kmeans_get_energy(VlKMeans *self)
    inline void* vl_kmeans_get_centers(VlKMeans *self)
    
    inline void vl_kmeans_set_algorithm(VlKMeans *self,
                                        VlKMeansAlgorithm algorithm)
    inline void vl_kmeans_set_initialization(
            VlKMeans *self, VlKMeansInitialization initialization)
    inline void vl_kmeans_set_num_repetitions(VlKMeans *self,
                                              vl_size numRepetitions)
    inline void vl_kmeans_set_max_num_iterations(VlKMeans *self,
                                                 vl_size maxNumIterations)
    inline void vl_kmeans_set_min_energy_variation(VlKMeans *self,
                                                   double minEnergyVariation)
    inline void vl_kmeans_set_verbosity(VlKMeans *self, int verbosity)
    inline void vl_kmeans_set_max_num_comparisons(VlKMeans *self,
                                                  vl_size maxNumComparisons)
    inline void vl_kmeans_set_num_trees(VlKMeans *self, vl_size numTrees)



# Copyright(C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from .host cimport vl_size, vl_uint32, vl_uint8, vl_int32, vl_uint
from .ikmeans cimport VlIKMFilt

cdef extern from "vl/hikmeans.h":
    cdef struct _VlHIKMNode:
        VlIKMFilt* filter  # IKM filter for this node*/
        _VlHIKMNode** children  # Node children (if any)
    ctypedef _VlHIKMNode VlHIKMNode

    cdef struct _VlHIKMTree:
        vl_size M  # IKM: data dimensionality
        vl_size K  # IKM: K
        vl_size depth  # Depth of the tree
        vl_size max_niters   # IKM: maximum # of iterations
        int method  # IKM: method
        int verb  # Verbosity level
        VlHIKMNode * root  # Tree root node
    ctypedef _VlHIKMTree VlHIKMTree

    VlHIKMTree *vl_hikm_new (int method)
    void vl_hikm_delete (VlHIKMTree *f)

    vl_size vl_hikm_get_ndims (VlHIKMTree *f)
    vl_size vl_hikm_get_K (VlHIKMTree *f)
    vl_size vl_hikm_get_depth (VlHIKMTree *f)
    int vl_hikm_get_verbosity (VlHIKMTree *f)
    vl_size vl_hikm_get_max_niters (VlHIKMTree *f)
    VlHIKMNode* vl_hikm_get_root (VlHIKMTree *f)

    void vl_hikm_set_verbosity (VlHIKMTree *f, int verb)
    void vl_hikm_set_max_niters (VlHIKMTree *f, int max_niters)

    void vl_hikm_init (VlHIKMTree *f, vl_size M, vl_size K, vl_size depth)
    void vl_hikm_train (VlHIKMTree *f, vl_uint8 *data, vl_size N)
    void vl_hikm_push (VlHIKMTree *f, vl_uint32 *asgn, vl_uint8 *data, vl_size N)


# Author: Alexis Mignon <alexis.mignon@probayes.com>
import numpy as np
import pytest
from numpy.testing import assert_allclose
from cyvlfeat.gmm import gmm

np.random.seed(1)
X = np.random.randn(1000, 2)
X[500:] *= (2, 3)
X[500:] += (4, 4)


def test_gmm_2_clusters_rand_init():
    means, covars, priors, LL, posteriors = gmm(X, n_clusters=2)

    assert_allclose(LL, -4341.0, atol=0.1)
    assert_allclose(priors, [0.5, 0.5], atol=0.1)
    assert_allclose(posteriors[0], [0.0, 1.0], atol=0.1)
    assert_allclose(means, [[4, 4], [0, 0]], atol=0.2)


def test_gmm_2_clusters_kmeans_init():
    means, covars, priors, LL, posteriors = gmm(X, n_clusters=2,
                                                init_mode='kmeans')

    assert_allclose(LL, -4341.0, atol=0.1)
    assert_allclose(priors, [0.5, 0.5], atol=0.1)
    assert_allclose(posteriors[0], [0.0, 1.0], atol=0.1)
    assert_allclose(means, [[4, 4], [0, 0]], atol=0.2)


def test_gmm_2_clusters_custom_init_fail():
    with pytest.raises(ValueError):
        _ = gmm(X, n_clusters=2, init_mode='custom')




# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from .host cimport vl_size, vl_bool

cdef extern from "vl/generic.h":
    void vl_set_simd_enabled(vl_bool x)
    vl_bool vl_get_simd_enabled()
    vl_bool vl_cpu_has_avx()
    vl_bool vl_cpu_has_sse3()
    vl_bool vl_cpu_has_sse2()
    vl_size vl_get_num_cpus()
    
    vl_size vl_get_max_threads()
    void vl_set_num_threads(vl_size n)
    vl_size vl_get_thread_limit()



# distutils: language = c
# Copyright (C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.
import numpy as np
cimport numpy as np
cimport cython
from cython.operator cimport dereference as deref
from libc.stdlib cimport qsort

# Import the header files
from cyvlfeat._vl.dsift cimport *
from cyvlfeat._vl.host cimport *
from cyvlfeat._vl.sift cimport *
from cyvlfeat._vl.mathop cimport VL_PI
from cyvlfeat.cy_util cimport py_printf, set_python_vl_printf


@cython.boundscheck(False)
cpdef cy_dsift(float[:, ::1] data, int[:] step,
               int[:] size, int[:] bounds, int window_size, bint norm,
               bint fast, bint float_descriptors, int[:] geometry,
               bint verbose):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        int num_frames = 0
        const VlDsiftKeypoint *frames_array
        const float *descriptors_array
        int k = 0, i = 0
        int step_x = 0, step_y = 0, min_x = 0, min_y = 0, max_x = 0, max_y = 0

        float[:, ::1] out_descriptors
        float[:, ::1] out_frames

        int height = data.shape[0]
        int width = data.shape[1]

        VlDsiftDescriptorGeometry geom
        VlDsiftFilter *dsift = vl_dsift_new(width, height)

        int ndims = 0
        int descriptor_index = 0
        float* linear_descriptor

    # Setup the geometry (number of bins and sizes)
    # Note the y-axis is taken as the first (zeroth) axis
    geom.numBinX = geometry[1]
    geom.numBinY = geometry[0]
    geom.numBinT = geometry[2]
    geom.binSizeX = size[1]
    geom.binSizeY = size[0]
    vl_dsift_set_geometry(dsift, &geom)

    # Set other options
    vl_dsift_set_steps(dsift, step[1], step[0])
    vl_dsift_set_bounds(dsift, bounds[1], bounds[0], bounds[3], bounds[2])
    vl_dsift_set_flat_window(dsift, fast)

    if window_size >= 0:
      vl_dsift_set_window_size(dsift, window_size)

    # Get calculated values from the dsift object
    num_frames = vl_dsift_get_keypoint_num(dsift)
    descriptor_length = vl_dsift_get_descriptor_size(dsift)
    geom = deref(vl_dsift_get_geometry(dsift))

    if verbose:
      vl_dsift_get_steps(dsift, &step_x, &step_y)
      vl_dsift_get_bounds(dsift, &min_x, &min_y, &max_x, &max_y)

      py_printf("vl_dsift: image size         [W, H] = [%d, %d]\n", width, height)
      py_printf("vl_dsift: bounds:            "
                "[minX, minY, maxX, maxY] = [%d, %d, %d, %d]\n",
                min_x, min_y, max_x, max_y)
      py_printf("vl_dsift: subsampling steps: stepX=%d, stepY=%d\n",
                step_x, step_y)
      py_printf("vl_dsift: num bins:          "
                "[numBinT, numBinX, numBinY] = [%d, %d, %d]\n",
                geom.numBinT,
                geom.numBinX,
                geom.numBinY)
      py_printf("vl_dsift: descriptor size:   %d\n", descriptor_length)
      py_printf("vl_dsift: bin sizes:         [binSizeX, binSizeY] = [%d, %d]\n",
                geom.binSizeX,
                geom.binSizeY)
      py_printf("vl_dsift: flat window:       %d\n", fast)
      py_printf("vl_dsift: window size:       "
                "%g\n", vl_dsift_get_window_size(dsift))
      py_printf("vl_dsift: num of features:   %d\n", num_frames)

    # Actually compute the SIFT features
    vl_dsift_process(dsift, &data[0, 0])

    # Grab the results
    frames_array = vl_dsift_get_keypoints(dsift)
    descriptors_array = vl_dsift_get_descriptors(dsift)

    # Create output arrays
    out_descriptors = np.empty((num_frames, descriptor_length),
                               dtype=np.float32, order='C')
    # Grab the pointer to the data so we can walk it linearly
    linear_descriptor = &out_descriptors[0, 0]

    # The norm is added as the third component if set
    if norm:
        ndims = 3
        out_frames = np.empty((num_frames, ndims), dtype=np.float32)
    else:
        ndims = 2
        out_frames = np.empty((num_frames, ndims), dtype=np.float32)

    # Copy results out
    for k in range(num_frames):
        out_frames[k, 0] = frames_array[k].y
        out_frames[k, 1] = frames_array[k].x

        # We have an implied / 2 in the norm, because of the clipping below
        if norm:
            out_frames[k, 2] = frames_array[k].norm

        # We don't need to transpose because our memory is in the correct
        # order already!
        for i in range(descriptor_length):
            descriptor_index = num_frames * i + k
            linear_descriptor[descriptor_index] = \
                min(512.0 * descriptors_array[descriptor_index], 255.0)

    # Clean up the allocated memory
    vl_dsift_delete(dsift)

    if float_descriptors:
        return np.asarray(out_frames), np.asarray(out_descriptors)
    else:
        return np.asarray(out_frames), np.asarray(out_descriptors).astype(np.uint8)


cdef int korder(const void *a, const void *b) nogil:
    cdef float x = (<float*> a)[2] - (<float*> b)[2]
    if x < 0: return -1
    if x > 0: return +1
    return 0


@cython.boundscheck(False)
cpdef cy_sift(float[:, ::1] data, int n_octaves,
              int n_levels, int first_octave, float peak_threshold,
              float edge_threshold, float norm_threshold, int magnification,
              int window_size, float[:, :] frames, bint force_orientations,
              bint float_descriptors, bint compute_descriptor, bint verbose):
    # Set the vlfeat printing function to the Python stdout
    set_python_vl_printf()

    cdef:
        bint is_first_octave = True
        int n_user_keypoints = 0, total_keypoints = 0, reserved = 0, i = 0, j = 0, q = 0
        int height = data.shape[0]
        int width = data.shape[1]
        float *user_keypoints_arr
        VlSiftFilt *filt = vl_sift_new(width, height, n_octaves, n_levels,
                                       first_octave)

        # Create empty 2D output arrays
        float[:, ::1] out_descriptors = np.empty((0, 128), dtype=np.float32,
                                                 order='C')
        float[:, ::1] out_frames = np.empty((0, 4), dtype=np.float32, order='C')

        float *flat_descriptors = &out_descriptors[0, 0]
        float *flat_out_frames = &out_frames[0, 0]

        int is_octaves_complete = 0
        const VlSiftKeypoint *keypoints
        int n_keypoints = 0

        double[4] angles
        int n_angles = 0
        const VlSiftKeypoint *curr_keypoint
        VlSiftKeypoint ik

        vl_sift_pix[128] single_descriptor_arr

        bint user_specified_frames = False

    user_specified_frames = frames is not None
    if user_specified_frames:
        n_user_keypoints = frames.shape[0]
        user_keypoints_arr = &frames[0, 0]
        # Ensure frames array is sorted by increasing scale
        qsort(user_keypoints_arr, n_user_keypoints,
              4 * sizeof(float), korder)

    if peak_threshold  >= 0: vl_sift_set_peak_thresh(filt, peak_threshold)
    if edge_threshold  >= 0: vl_sift_set_edge_thresh(filt, edge_threshold)
    if norm_threshold  >= 0: vl_sift_set_norm_thresh(filt, norm_threshold)
    if magnification   >= 0: vl_sift_set_magnif(filt, magnification)
    if window_size     >= 0: vl_sift_set_window_size(filt, window_size)

    if verbose:
        py_printf("vl_sift: filter settings:\n")
        py_printf("vl_sift:   octaves      (O)      = %d\n", vl_sift_get_noctaves(filt))
        py_printf("vl_sift:   levels       (S)      = %d\n", vl_sift_get_nlevels(filt))
        py_printf("vl_sift:   first octave (o_min)  = %d\n", vl_sift_get_octave_first(filt))
        py_printf("vl_sift:   edge thresh           = %g\n", vl_sift_get_edge_thresh(filt))
        py_printf("vl_sift:   peak thresh           = %g\n", vl_sift_get_peak_thresh(filt))
        py_printf("vl_sift:   norm thresh           = %g\n", vl_sift_get_norm_thresh(filt))
        py_printf("vl_sift:   window size           = %g\n", vl_sift_get_window_size(filt))
        py_printf("vl_sift:   float descriptor      = %d\n", float_descriptors)

        py_printf("vl_sift: will source frames? yes (%d read)\n"
                  if user_specified_frames
                  else "vl_sift: will source frames? no\n", n_user_keypoints)
        py_printf("vl_sift: will force orientations? %d\n", force_orientations)


    if user_specified_frames:
        # If we have specified the frames, we know how many frames
        # will be calculated, so we can skip the dynamic reallocation
        # which is normally done inside the keypoints loop
        out_frames = np.resize(out_frames, (n_user_keypoints, 4))
        flat_out_frames = &out_frames[0, 0]
        # Similar for the descriptors, if necessary
        if compute_descriptor:
            out_descriptors = np.resize(out_descriptors,
                                        (n_user_keypoints, 128))
            flat_descriptors = &out_descriptors[0, 0]

    # Process each octave
    while True:
        if verbose:
            py_printf("vl_sift: processing octave %d\n",
                      vl_sift_get_octave_index(filt))

        # Calculate the GSS for the next octave ....................
        if is_first_octave:
            is_octaves_complete = vl_sift_process_first_octave(filt,
                                                               &data[0, 0])
            is_first_octave = False
        else:
            is_octaves_complete = vl_sift_process_next_octave(filt)

        if is_octaves_complete:
            break

        if verbose:
            py_printf("vl_sift: GSS octave %d computed\n",
                      vl_sift_get_octave_index(filt))

        # Run detector .............................................
        if not user_specified_frames:
            i = 0
            vl_sift_detect(filt)

            keypoints = vl_sift_get_keypoints(filt)
            n_keypoints = vl_sift_get_nkeypoints(filt)

            if verbose:
                py_printf("vl_sift: detected %d (unoriented) keypoints\n",
                          n_keypoints)
        else:
            n_keypoints = n_user_keypoints

        # For each keypoint
        while i < n_keypoints:
            # Obtain keypoint orientations
            if user_specified_frames:
                vl_sift_keypoint_init(filt, &ik,
                                      user_keypoints_arr[4 * i + 1],
                                      user_keypoints_arr[4 * i + 0],
                                      user_keypoints_arr[4 * i + 2])

                if ik.o != vl_sift_get_octave_index(filt):
                    break

                curr_keypoint = &ik

                # Optionally force computation of orientations
                if force_orientations:
                    n_angles = vl_sift_calc_keypoint_orientations(filt, angles,
                                                                  curr_keypoint)
                else:
                    angles[0] = user_keypoints_arr[4 * i + 3]
                    n_angles  = 1
            else:
                # This is equivalent to &keypoints[i] - just get the pointer
                # to the i'th element.
                curr_keypoint = keypoints + i
                n_angles = vl_sift_calc_keypoint_orientations(filt, angles,
                                                              curr_keypoint)

            # For each orientation
            for q in range(n_angles):
                if compute_descriptor:
                    vl_sift_calc_keypoint_descriptor(filt,
                                                     single_descriptor_arr,
                                                     curr_keypoint, angles[q])

                # Dynamically reallocate the output arrays so that they can
                # fit all the keypoints being requested.
                # If statement says: IF we will run out of space next iteration
                #                    AND we have computed the frame OR the user
                #                        has allowed estimation of the number of
                #                        orientations AND there was more than one
                #                    THEN reallocate memory
                if (reserved < total_keypoints + 1 and
                   (not user_specified_frames or
                    (force_orientations and n_angles > 1))):
                    reserved += 2 * n_keypoints

                    out_frames = np.resize(out_frames, (reserved, 4))
                    flat_out_frames = &out_frames[0, 0]

                    if compute_descriptor:
                        out_descriptors = np.resize(out_descriptors,
                                                    (reserved, 128))
                        flat_descriptors = &out_descriptors[0, 0]

                # Notice that this method will give different results
                # from MATLAB because MATLAB actually runs on the
                # transpose of the image due to it's fortran ordering!
                flat_out_frames[total_keypoints * 4 + 0] = curr_keypoint.y
                flat_out_frames[total_keypoints * 4 + 1] = curr_keypoint.x
                flat_out_frames[total_keypoints * 4 + 2] = curr_keypoint.sigma
                flat_out_frames[total_keypoints * 4 + 3] = angles[q]

                if compute_descriptor:
                    for j in range(128):
                        flat_descriptors[total_keypoints * 128 + j] = \
                            min(512.0 * single_descriptor_arr[j], 255.0)

                total_keypoints += 1
            i += 1

    if verbose:
        py_printf("vl_sift: found %d keypoints\n", total_keypoints)

    # cleanup
    vl_sift_delete(filt)

    # If we have dynamically allocated memory for the frames, make sure that
    # we resize the array back to the correct size (since we optimistically
    # allocated previously to reduce the number of total resizes)
    if out_frames.shape[0] != total_keypoints:
        out_frames = np.resize(out_frames, (total_keypoints, 4))
        out_descriptors = np.resize(out_descriptors, (total_keypoints, 128))

    if compute_descriptor:
        if float_descriptors:
            return np.asarray(out_frames), np.asarray(out_descriptors)
        else:
            return np.asarray(out_frames), np.asarray(out_descriptors).astype(np.uint8)
    else:
        return np.asarray(out_frames)



# Copyright(C) 2007-12 Andrea Vedaldi and Brian Fulkerson.
# All rights reserved.

# This file is modified from part of the VLFeat library and is made available
# under the terms of the BSD license.

from .host cimport vl_size


cdef extern from "vl/sift.h":
    ctypedef float vl_sift_pix

    ctypedef struct VlSiftKeypoint:
        int o          # < o coordinate(octave).
        int ix         # < Integer unnormalized x coordinate.
        int iy         # < Integer unnormalized y coordinate.
        int c_is "is"  # < Integer s coordinate.
        float x        # < x coordinate.
        float y        # < y coordinate.
        float s        # < s coordinate.
        float sigma    # < scale.

    ctypedef struct VlSiftFilt:
        double sigman        # < nominal image smoothing. 
        double sigma0        # < smoothing of pyramid base. 
        double sigmak        # < k-smoothing 
        double dsigma0       # < delta-smoothing.
        int width            # < image width. 
        int height           # < image height. 
        int O                # < number of octaves. 
        int S                # < number of levels per octave. 
        int o_min            # < minimum octave index. 
        int s_min            # < minimum level index. 
        int s_max            # < maximum level index. 
        int o_cur            # < current octave.
        vl_sift_pix *temp    # < temporary pixel buffer. 
        vl_sift_pix *octave  # < current GSS data. 
        vl_sift_pix *dog     # < current DoG data. 
        int octave_width     # < current octave width. 
        int octave_height    # < current octave height.
        vl_sift_pix *gaussFilter   # < current Gaussian filter 
        double gaussFilterSigma    # < current Gaussian filter std 
        vl_size gaussFilterWidth   # < current Gaussian filter width
        VlSiftKeypoint* keys # < detected keypoints. 
        int nkeys            # < number of detected keypoints. 
        int keys_res         # < size of the keys buffer.
        double peak_thresh   # < peak threshold. 
        double edge_thresh   # < edge threshold. 
        double norm_thresh   # < norm threshold. 
        double magnif        # < magnification factor. 
        double windowSize    # < size of Gaussian window(in spatial bins)
        vl_sift_pix *grad    # < GSS gradient data. 
        int grad_o           # < GSS gradient data octave.

    VlSiftFilt* vl_sift_new(int width, int height, int noctaves, int nlevels,
                            int o_min)
    void vl_sift_delete(VlSiftFilt *f) 
    int vl_sift_process_first_octave(VlSiftFilt *f, vl_sift_pix *im)
    int vl_sift_process_next_octave(VlSiftFilt *f) 
    void vl_sift_detect(VlSiftFilt *f) 
    int vl_sift_calc_keypoint_orientations(VlSiftFilt *f, double *angles,
                                            VlSiftKeypoint *k)
    void vl_sift_calc_keypoint_descriptor(VlSiftFilt *f, vl_sift_pix *descr,
                                          VlSiftKeypoint * k, double angle)
    void vl_sift_calc_raw_descriptor(VlSiftFilt *f,
                                     vl_sift_pix * image,
                                     vl_sift_pix *descr, int width, int height,
                                     double x, double y, double s,
                                     double angle0)
    void vl_sift_keypoint_init(VlSiftFilt *f, VlSiftKeypoint *k, double x,
                               double y, double sigma)
    inline int vl_sift_get_octave_index(VlSiftFilt *f)
    inline int vl_sift_get_noctaves(VlSiftFilt *f)
    inline int vl_sift_get_octave_first(VlSiftFilt *f)
    inline int vl_sift_get_octave_width(VlSiftFilt *f)
    inline int vl_sift_get_octave_height(VlSiftFilt *f)
    inline int vl_sift_get_nlevels(VlSiftFilt *f)
    inline int vl_sift_get_nkeypoints(VlSiftFilt *f)
    inline double vl_sift_get_peak_thresh(VlSiftFilt *f)
    inline double vl_sift_get_edge_thresh(VlSiftFilt *f)
    inline double vl_sift_get_norm_thresh(VlSiftFilt *f)
    inline double vl_sift_get_magnif(VlSiftFilt *f)
    inline double vl_sift_get_window_size(VlSiftFilt *f)
    inline vl_sift_pix *vl_sift_get_octave  (VlSiftFilt *f, int s)
    inline VlSiftKeypoint *vl_sift_get_keypoints (VlSiftFilt *f)
    inline void vl_sift_set_peak_thresh(VlSiftFilt *f, double t) 
    inline void vl_sift_set_edge_thresh(VlSiftFilt *f, double t) 
    inline void vl_sift_set_norm_thresh(VlSiftFilt *f, double t) 
    inline void vl_sift_set_magnif(VlSiftFilt *f, double m) 
    inline void vl_sift_set_window_size(VlSiftFilt *f, double m) 
